/***************************************************************************
 *   Copyright (C) 2007 by Jan Vermeulen                                   *
 *   janverm@gmail.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "liquid.h"

typedef struct {
	int x, y;
} bump_t;

static short aSin1[512];
static short aSin2[512];
static float reflectionmap[256][256];
static bump_t bumpmap[RES_X][RES_Y];
static Uint16 sin_index = 0;
static Uint16 sin_index2 = 80;
static SDL_Surface *picture = NULL;
static SDL_Surface *s = NULL;
static SDL_Surface *heightmap_rgb[3];
static SDL_Surface *bump_effect = NULL;
static SDL_Surface *greypicture = NULL;
static Uint16 lightx, lighty, vx, vy;
static short normalx, normaly, x, y;
static int inited = 0;
static int temp;

static Uint16 adjust_color(Uint16 color, Uint8 gradient)
{
	temp = color;

	/* color must be RGB565 */
	Uint16 r,g,b,c;

#ifdef MULTICOLOR
	/* strip colours */
	r = ((color & 0xf800) >> 11);
	g = ((color & 0x07e0) >> 5);
	b = ((color & 0x001f));

	/* apply gradient, 
	 * 0   = black
	 * 128 = original color
	 * 256 = highlighted
	 */
	r += (Uint8)(gradient*1.1); 
	g += (Uint8)(gradient*1.1);
	b += (Uint8)(gradient*1.1);

	/* clamp values */
	if (r > 255) r = 255;
	if (g > 255) g = 255;
	if (b > 255) b = 255;
#endif	
	/* combine again  */
	r = g = b = gradient; 
	c = (((r & 0xf8) << 8) |
	     ((g & 0xfc) << 3) |
	     ((b & 0xf8) >> 3));

	return c;
}

static void create_heightmap(SDL_Surface * surf)
{
	SDL_Palette *p;
	Uint16 i, j;
	int x;
	Uint8 index;
	Uint8 *image;

	if (!surf->format->palette) {
		/* not an 8-bit surf surface */
		printf("Wow ... not an 8-bit surface!");
		return;
	}

	p = surf->format->palette;

	/* sort palette from dark to light colors using selection sort */
	for (i = 0; i < 256; ++i) {
		SDL_Color *c = p->colors + i;
		index = i;

		for (j = i + 1; j < 256; ++j) {
			SDL_Color *d = p->colors + j;
			Uint8 biggest = 0;

			/* what is the biggest value? */
			if (c->r >= c->g && c->r >= c->b) {
				/* r is biggest */
				biggest = c->r;
			} else if (c->g >= c->r && c->g >= c->b) {
				/* g is biggest */
				biggest = c->g;
			} else if (c->b >= c->r && c->b >= c->g) {
				/* b is biggest */
				biggest = c->b;
			}

			if (d->r <= biggest && d->g <= biggest
			    && d->b <= biggest) {
				if (d->r == biggest || d->g == biggest
				    || d->b == biggest) {
					if (d->r + d->g + d->b <
					    c->r + c->g + c->b) {
						/* found darker color */
						index = j;
						c = d;
					}
				} else {
					/* found darker color */
					index = j;
					c = d;
				}
			}
		}

		image = (Uint8 *) surf->pixels;
		p->colors[index] = p->colors[i];
		p->colors[i].r = p->colors[i].g = p->colors[i].b = i;

		/* swap pixels in image */
		for (x = 0; x < surf->w * surf->h; ++x) {
			if (*image == index)
				*image = i;
			image++;
		}
	}

}

static void split_image_rgb_grey(SDL_Surface * source, SDL_Surface * rgb[],
				 SDL_Surface * greysurf)
{
	/* this function splits a RGB 'source' surface into 3 
	 * surfaces which compose only its red, green and blue 
	 * component, and an image with 8b greyscale image.
	 */
	SDL_Color col[256];
	Uint8 r, g, b;
	int i;
	Uint16 pxl;
	Uint16 *src = (Uint16 *) source->pixels;
	Uint8 *red = (Uint8 *) rgb[0]->pixels;
	Uint8 *green = (Uint8 *) rgb[1]->pixels;
	Uint8 *blue = (Uint8 *) rgb[2]->pixels;
	Uint8 *grey = (Uint8 *) greysurf->pixels;

	for (i = 0; i < 256; i++) {
		col[i].r = col[i].g = col[i].b = i;
	}

	for (i = 0; i < (source->w * source->h); i++) {
		pxl = *src++;
		r = (Uint8) ((pxl & 0xf800) >> 11);
		g = (Uint8) ((pxl & 0x07e0) >> 5);
		b = (Uint8) ((pxl & 0x001f));
		*red++ = r;
		*green++ = g;
		*blue++ = b;
		*grey++ = (Uint8) ((r + g + b) / 3);
	}

	SDL_SetPalette(greysurf, SDL_LOGPAL | SDL_PHYSPAL, col, 0, 256);

	for (i = 0; i < 3; i++)
		SDL_SetPalette(rgb[i], SDL_LOGPAL | SDL_PHYSPAL, col, 0, 256);
}

int fx_bump_init(unsigned long val)
{
	SDL_Surface *screen = (SDL_Surface *) val;
	Uint16 i;
	int x, y;
	float rad;

	if (inited) return 0;
	
	bump_effect = SDL_CreateRGBSurface(SDL_SWSURFACE,
					   RES_X, RES_Y, 16,
					   screen->format->Rmask,
					   screen->format->Gmask,
					   screen->format->Bmask,
					   screen->format->Amask);

#if (RES_X == 480)
	s = SDL_LoadBMP("./pix/chamwoman.bmp");
#elif (RES_X == 800)
	s = SDL_LoadBMP("./pix/chamwoman_hi.bmp");
#else
#error Only horizontal resolutions of 480 or 800 are supported.
#endif

	picture = SDL_DisplayFormat(s);
	if (!picture) {
		fprintf(stderr, "Cannot open chamwoman picture: %s\n",
			SDL_GetError());
		return -1;
	}

	/* red */
	heightmap_rgb[0] = SDL_CreateRGBSurface(SDL_SWSURFACE,
						picture->w, picture->h, 8,
						0, 0, 0, 0);
	/* green */
	heightmap_rgb[1] = SDL_CreateRGBSurface(SDL_SWSURFACE,
						picture->w, picture->h, 8,
						0, 0, 0, 0);
	/* blue */
	heightmap_rgb[2] = SDL_CreateRGBSurface(SDL_SWSURFACE,
						picture->w, picture->h, 8,
						0, 0, 0, 0);
	/* grey */
	greypicture = SDL_CreateRGBSurface(SDL_SWSURFACE,
					   picture->w, picture->h, 8,
					   0, 0, 0, 0);

	split_image_rgb_grey(picture, heightmap_rgb, greypicture);

	/* create sin lookup table */
	for (i = 0; i < 512; i++) {
		rad = (float)i *0.0174532 * 0.703125;
		aSin1[i] = (short)((sin(rad) * 240.0) + 320.0);
		aSin2[i] = (short)((sin(rad) * 100.0) + 256.0);
	}

	/* create reflection map */
	for (x = 0; x < 256; ++x) {
		for (y = 0; y < 256; ++y) {
			float X = (x - 128) / 128.0;
			float Y = (y - 128) / 128.0;
			float Z = 1.0 - sqrt(X * X + Y * Y);
			Z *= 255.0;
			if (Z < 0.0)
				Z = 0.0;
			reflectionmap[x][y] = Z;
		}
	}

	create_heightmap(greypicture);

	/* create bump map */
	for (x = 1; x < RES_X - 1; ++x) {
	for (y = 1; y < RES_Y - 1; ++y) {
		bumpmap[x][y].x = *((Uint8 *) greypicture->pixels + 
				   y * greypicture->pitch + x + 1)
				   - *((Uint8 *) greypicture->pixels +
				   y * greypicture->pitch + x - 1);
		bumpmap[x][y].y = *((Uint8 *) greypicture->pixels +
			          ((y + 1) * greypicture->pitch) + x)
			          - *((Uint8 *) greypicture->pixels +
				  ((y - 1) * greypicture->pitch) + x);
	}
	}

	inited = 1;
	return 0;
}

int fx_bump_exit(unsigned long val)
{
	temp = val;

	SDL_FreeSurface(heightmap_rgb[0]);
	SDL_FreeSurface(heightmap_rgb[1]);
	SDL_FreeSurface(heightmap_rgb[2]);
	SDL_FreeSurface(greypicture);
	SDL_FreeSurface(picture);
	SDL_FreeSurface(s);
	SDL_FreeSurface(bump_effect);
	return 0;
}

SDL_Surface *fx_bump_play(unsigned long val)
{
	Uint16 *s1;
	Uint16 *s2;

	temp = val;

	lightx = aSin1[sin_index];
	lighty = aSin2[sin_index2];

	s1 = (Uint16 *) bump_effect->pixels + bump_effect->pitch + 1;
	s2 = (Uint16 *) picture->pixels + picture->pitch + 1;
	
	for (y = 1; y < bump_effect->h - 1; ++y) {
		vy = lighty - y;
		for (x = 1; x < bump_effect->w - 1; ++x) {
				vx = lightx - x;
				normalx = bumpmap[x][y].x + vx;
				normaly = bumpmap[x][y].y + vy;

				if (normalx < 0)
					normalx = 0;
				else if (normalx > 255)
					normalx = 0;
				if (normaly < 0)
					normaly = 0;
				else if (normaly > 255)
					normaly = 0;
		
				*s1++ = adjust_color(*s2++, (Uint8)reflectionmap[normalx][normaly]);
		}
		s1+=2;
		s2+=2;
	}

	sin_index += 3;
	sin_index &= 511;
	sin_index2 += 5;
	sin_index2 &= 511;
#if 0
	SDL_Delay(20);
#endif
	return bump_effect;
}

struct demopart_t fx_bump = {
	.time = 15,
	.init = fx_bump_init,
	.exit = fx_bump_exit,
	.play = fx_bump_play,
	.info = "Bump mapping!",
};
