/***************************************************************************
 *   Copyright (C) 2007 by Jan Vermeulen                                   *
 *   janverm@gmail.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "liquid.h"

#define LENS_WIDTH 150
#define LENS_ZOOM  30
#define LENS_OFFSET 2

static SDL_Surface *lens_effect = NULL;
static SDL_Surface *backgnd = NULL;
static Uint32 lens[LENS_WIDTH][LENS_WIDTH];
static int x = 16, y = 16; 
static int xd = 1, yd = 1;
static int inited = 0;
static int temp;

static void apply_lens(int ox, int oy)
{
	Uint16 *image = (Uint16 *) lens_effect->pixels;
	Uint16 *back = (Uint16 *) backgnd->pixels;
	int x, y, temp, pos;
	for (y = 0; y < LENS_WIDTH; y++) {
		temp = (y + oy) * RES_X + ox;
		for (x = 0; x < LENS_WIDTH; x++) {
			pos = temp + x;
			image[pos] = back[pos + lens[y][x]];
		}
	}
}

int fx_lens_init(unsigned long val)
{
	int x, y, r, d;
	SDL_Surface *s;
	SDL_Surface *screen = (SDL_Surface *) val;

	if (inited) return 0;
	
	lens_effect = SDL_CreateRGBSurface(SDL_SWSURFACE,
					   RES_X, RES_Y, 16,
					   screen->format->Rmask,
					   screen->format->Gmask,
					   screen->format->Bmask,
					   screen->format->Amask);

	/* load the background image */
#if (RES_X == 480)
	s = SDL_LoadBMP("./pix/chamwoman.bmp");
#elif (RES_X == 800)
	s = SDL_LoadBMP("./pix/chamwoman_hi.bmp");
#else
#error Only horizontal resolutions of 480 or 800 are supported.
#endif

	backgnd = SDL_DisplayFormat(s);
	if (!backgnd) {
		fprintf(stderr, "Cannot open file chamwoman.bmp: %s\n",
			SDL_GetError());
		return -1;
	}

	/* generate the lens distortion */
	r = LENS_WIDTH / 2;
	d = LENS_ZOOM;

	/* the shift in the following expression is a function of the
	 * distance of the current point from the center of the sphere.
	 * If you imagine:
	 *
	 *       eye
	 *
	 *   .-~~~~~~~-.    sphere surface
	 * .`           '.
	 * ---------------  viewing plane
	 *        .         center of sphere
	 *
	 * For each point across the viewing plane, draw a line from the
	 * point on the sphere directly above that point to the center of
	 * the sphere.  It will intersect the viewing plane somewhere
	 * closer to the center of the sphere than the original point.
	 * The shift function below is the end result of the above math,
	 * given that the height of the point on the sphere can be derived
	 * from:
	 *
	 * x^2 + y^2 + z^2 = radius^2
	 *
	 * x and y are known, z is based on the height of the viewing
	 * plane.
	 *
	 * The radius of the sphere is the distance from the center of the
	 * sphere to the edge of the viewing plane, which is a neat little
	 * triangle.  If d = the distance from the center of the sphere to
	 * the center of the plane (aka, LENS_ZOOM) and r = half the width
	 * of the plane (aka, LENS_WIDTH/2) then radius^2 = d^2 + r^2.
	 *
	 * Things become simpler if we take z=0 to be at the plane's
	 * height rather than the center of the sphere, turning the z^2 in
	 * the expression above to (z+d)^2, since the center is now at
	 * (0, 0, -d).
	 *
	 * So, the resulting function looks like:
	 *
	 * x^2 + y^2 + (z+d)^2 = d^2 + r^2
	 *
	 * Expand the (z-d)^2:
	 *
	 * x^2 + y^2 + z^2 + 2dz + d^2 = d^2 + r^2
	 *
	 * Rearrange things to be a quadratic in terms of z:
	 *
	 * z^2 + 2dz + x^2 + y^2 - r^2 = 0
	 *
	 * Note that x, y, and r are constants, so apply the quadratic
	 * formula:
	 *
	 * For ax^2 + bx + c = 0,
	 * 
	 * x = (-b +- sqrt(b^2 - 4ac)) / 2a
	 *
	 * We can ignore the negative result, because we want the point at
	 * the top of the sphere, not at the bottom.
	 *
	 * x = (-2d + sqrt(4d^2 - 4 * (x^2 + y^2 - r^2))) / 2
	 *
	 * Note that you can take the -4 out of both expressions in the
	 * square root to put -2 outside, which then cancels out the
	 * division:
	 *
	 * z = -d + sqrt(d^2 - (x^2 + y^2 - r^2))
	 *
	 * This now gives us the height of the point on the sphere
	 * directly above the equivalent point on the plane.  Next we need
	 * to find where the line between this point and the center of the
	 * sphere at (0, 0, -d) intersects the viewing plane at (?, ?, 0).
	 * This is a matter of the ratio of line below the plane vs the
	 * total line length, multiplied by the (x,y) coordinates.  This
	 * ratio can be worked out by the height of the line fragment
	 * below the plane, which is d, and the total height of the line,
	 * which is d + z, or the height above the plane of the sphere
	 * surface plus the height of the plane above the center of the
	 * sphere.
	 *
	 * ratio = d/(d + z)
	 *
	 * Subsitute in the formula for z:
	 *
	 * ratio = d/(d + -d + sqrt(d^2 - (x^2 + y^2 - r^2))
	 *
	 * Simplify to:
	 *
	 * ratio = d/sqrt(d^2 - (x^2 + y^2 - r^2))
	 *
	 * Since d and r are constant, we now have a formula we can apply
	 * for each (x,y) point within the sphere to give the (x',y')
	 * coordinates of the point we should draw to project the image on
	 * the plane to the surface of the sphere.  I subtract the
	 * original (x,y) coordinates to give an offset rather than an
	 * absolute coordinate, then convert that offset to the image
	 * dimensions, and store the offset in a matrix the size of the
	 * intersecting circle.  Drawing the lens is then a matter of:
	 *
	 * screen[coordinate] = image[coordinate + lens[y][x]]
	 *
	 */

	/* it is sufficient to generate 1/4 of the lens and reflect this
	 * around; a sphere is mirrored on both the x and y axes */
	for (y = 0; y < LENS_WIDTH >> 1; y++) {
		for (x = 0; x < LENS_WIDTH >> 1; x++) {
			int ix, iy, offset;
			if ((x * x + y * y) < (r * r)) {
				float shift =
				    d / sqrt(d * d - (x * x + y * y - r * r));
				ix = x * shift - x;
				iy = y * shift - y;
			} else {
				ix = 0;
				iy = 0;
			}
			offset = (iy * RES_X + ix);
			lens[LENS_WIDTH / 2 - y][LENS_WIDTH / 2 - x] = -offset;
			lens[LENS_WIDTH / 2 + y][LENS_WIDTH / 2 + x] = offset;
			offset = (-iy * RES_X + ix);
			lens[LENS_WIDTH / 2 + y][LENS_WIDTH / 2 - x] = -offset;
			lens[LENS_WIDTH / 2 - y][LENS_WIDTH / 2 + x] = offset;
		}
	}

	SDL_FreeSurface(s);
	inited = 1;
	return 0;
}

int fx_lens_exit(unsigned long val)
{
	temp = val;
	SDL_FreeSurface(lens_effect);
	return 0;
}

SDL_Surface *fx_lens_play(unsigned long val)
{
	temp = val;

	/* blit the whole backing image across */
	SDL_BlitSurface(backgnd, NULL, lens_effect, NULL);

	/* apply the lens */
	apply_lens(x, y);

	/* shift the coordinates around */
	x += xd;
	y += yd;
	if (x > (RES_X - LENS_WIDTH - LENS_OFFSET) || x < LENS_OFFSET)
		xd = -xd;
	if (y > (RES_Y - LENS_WIDTH - LENS_OFFSET) || y < LENS_OFFSET)
		yd = -yd;

	return lens_effect;
}

struct demopart_t fx_lens = {
	.time = 10,
	.init = fx_lens_init,
	.exit = fx_lens_exit,
	.play = fx_lens_play,
	.info = "Lens effect!",
};
